import numpy as np
from socialAD.forward import forwardAD

def gradient_descent(variables, f, cur_x, rate, precision, previous_step_size, max_iters):
    if isinstance(f, str):

    #If just one string, convert to list
        if isinstance(variables, str):

            variables = [variables]

        #Convert to numpy arrays
        variables = np.array(variables)
        cur_x = np.array(cur_x)

        iters = 0
        while previous_step_size > precision and iters < max_iters:
            for index, variable in enumerate(variables):

                #Store to variable name
                name = variable
                #Save as forwardAD object
                exec(name + " = forwardAD(np.array(cur_x[index]), numvar = len(variables), idx = index)")
            #Stores full function as a forwardAD object
            AD_function = eval(f)
            
            prev_x = cur_x #Store current x value in prev_x
            cur_x = cur_x - rate * AD_function.der #Grad descent
            previous_step_size = np.sum(abs(cur_x - prev_x)) #Change in x
            iters += 1 #iteration count
            
    print("Iteration",iters) #,"\nX value is", cur_x) #Print iterations
    print("The local minimum occurs at", cur_x)


def gd_backtrack(variables, f, cur_x, precision, previous_step_size, max_iters):
    if isinstance(f, str):

    #If just one string, convert to list
        if isinstance(variables, str):

            variables = [variables]

        #Convert to numpy arrays
        variables = np.array(variables)
        cur_x = np.array(cur_x)
        
        f_x = f
        
        if len(variables) > 1:    
            
            for idx_var in np.arange(0, len(variables)):
                f_x = f_x.replace(variables[idx_var], variables[0] + '[' + str(idx_var) + ']')
        to_be_eval = "lambda " + variables[0] + " : " + f_x
        f_ex = eval( to_be_eval)
#        print(f_ex(cur_x))
#######################################
# ITERATIONS >>>>>>>>>>>>>>>>>>>>>>>>>

        iters = 0
        while previous_step_size > precision and iters < max_iters:
            for index, variable in enumerate(variables):

                #Store to variable name
                name = variable

                #Save as forwardAD object
                exec(name + " = forwardAD(np.array(cur_x[index]), numvar = len(variables), idx = index)")
#                print(eval(name))
                
            #Stores full function as a forwardAD object
            AD_function = eval(f)
            
            prev_x = cur_x #Store current x value in prev_x
            ##### ESTIMATE THE MOST EFFICIENT RATE 
            t = 1
            count = 1
            alpha = 0.3
            beta = 0.8
            while (AD_function.val - (f_ex(cur_x - t* AD_function.der) + alpha * t * np.dot( AD_function.der,  AD_function.der))) < 0 and count < max_iters:
                t *= beta

                count += 1
            
            rate = t
            #####
            
            cur_x = cur_x - rate * AD_function.der #Grad descent
            previous_step_size = np.sum(abs(cur_x - prev_x))  #Change in x
            iters += 1 #iteration count
    #        print("Iteration",iters,"\nX value is", cur_x) #Print iterations
#######################################
    print("Iteration",iters)
    print("The local minimum occurs at", cur_x)
    
# BFGS

def BFGS(variables, f, cur_x, precision, max_iters):

    if isinstance(f, str):
    #If just one string, convert to list
        if isinstance(variables, str):

            variables = [variables]

        #Convert to numpy arrays
        variables = np.array(variables)
        cur_x = np.array(cur_x) # points = np.array([2, 2])
        f_x = f
        
        if len(variables) > 1:    
            
            for idx_var in np.arange(0, len(variables)):
                f_x = f_x.replace(variables[idx_var], variables[0] + '[' + str(idx_var) + ']')
    

        to_be_eval = "lambda " + variables[0] + " : " + f_x
        f_ex = eval( to_be_eval)
#######################################        
        previous_step_size = 1
        p = 1e-8
        I = np.identity(len(variables))
        Hk = I
#######################################
# ITERATIONS >>>>>>>>>>>>>>>>>>>>>>>>>

        rate = 0.1
        iters = 0
        while previous_step_size > precision and iters < max_iters:
            for index, variable in enumerate(variables):

                #Store to variable name
                name = variable

                #Save as forwardAD object
                exec(name + " = forwardAD(np.array(cur_x[index]), numvar = len(variables), idx = index)")

            #Stores full function as a forwardAD object
            AD_function = eval(f)
            prev_x = cur_x
            
            neg_f_xk =  - AD_function.der
            s_k = Hk @ neg_f_xk * 0.0001 # with fix learning rate

            cur_x = prev_x + s_k
            y_k = f_ex(cur_x) + neg_f_xk
    
            rho_k = 1/(np.transpose(y_k) @ s_k)
            new_Hk = (I - s_k * rho_k @ np.transpose(y_k)) \
            @ Hk @ (I - rho_k * y_k@np.transpose(s_k))\
            + rho_k * s_k@np.transpose(s_k)
            Hk = new_Hk

            prev_x = cur_x #Store current x value in prev_x
            ##### ESTIMATE THE MOST EFFICIENT RATE 
            #####
            #print(rate)
            #print(AD_function.der)
            cur_x = cur_x - rate * AD_function.der #Grad descent
            previous_step_size = np.sum(abs(cur_x - prev_x)) #Change in x
            iters += 1 #iteration count
            #print("Iteration",iters,"\nX value is", cur_x) #Print iterations
#######################################
    print("Iteration",iters)
    print("The local minimum occurs at", cur_x)
    
    
    
# BFGS

def BFGS_backtrack(variables, f, cur_x, precision, max_iters):
    



    if isinstance(f, str):
    #If just one string, convert to list
        if isinstance(variables, str):

            variables = [variables]

        #Convert to numpy arrays
        variables = np.array(variables)
        cur_x = np.array(cur_x) # points = np.array([2, 2])
       
        f_x = f
        
        if len(variables) > 1:    
            
            for idx_var in np.arange(0, len(variables)):
                f_x = f_x.replace(variables[idx_var], variables[0] + '[' + str(idx_var) + ']')
    
        to_be_eval = "lambda " + variables[0] + " : " + f_x
        f_ex = eval( to_be_eval)
#######################################        
        previous_step_size = 1
        p = 1e-8
        I = np.identity(len(variables))
        Hk = I
#######################################
# ITERATIONS >>>>>>>>>>>>>>>>>>>>>>>>>

        rate = 0.1
        iters = 0
        while previous_step_size > precision and iters < max_iters:
            for index, variable in enumerate(variables):

                #Store to variable name
                name = variable

                #Save as forwardAD object
                exec(name + " = forwardAD(np.array(cur_x[index]), numvar = len(variables), idx = index)")

            #Stores full function as a forwardAD object
            #print(f)
            AD_function = eval(f) ## prev step 
            prev_x = cur_x
            
            neg_f_xk =  - AD_function.der
            s_k = Hk @ neg_f_xk * 0.0001 # with fix learning rate

            cur_x = prev_x + s_k
            
            #######################
            ## backtrack_AD
            #######################
#            t = backtrack_AD(points, 1, alpha, beta, 1)
            
            t = 1
            count = 1
           # print('=============')
           # print(AD_function.val) #<-- SHOULD BE ONE
           # print(cur_x )
           # print(- t* AD_function.der)
            alpha = 0.3
            beta = 0.8
            while (AD_function.val - (f_ex(cur_x - t* AD_function.der) + alpha * t * np.dot( AD_function.der,  AD_function.der))) < 0 and count < max_iters:
                t *= beta
                count += 1
            
            #print(f't: {t}')
            rate = t
            ######################
            y_k = f_ex(cur_x) + neg_f_xk * t 
    
            rho_k = 1/(np.transpose(y_k) @ s_k)
            new_Hk = (I - s_k * rho_k @ np.transpose(y_k)) \
            @ Hk @ (I - rho_k * y_k@np.transpose(s_k))\
            + rho_k * s_k@np.transpose(s_k)
            Hk = new_Hk

            prev_x = cur_x #Store current x value in prev_x
            ##### ESTIMATE THE MOST EFFICIENT RATE 
            #####
            #print(AD_function.der)
            cur_x = cur_x - rate * AD_function.der #Grad descent
            previous_step_size =  np.sum(abs(cur_x - prev_x))  #Change in x
            iters += 1 #iteration count
            #print("Iteration",iters,"\nX value is", cur_x) #Print iterations
#######################################
    print("Iteration",iters)
    print("The local minimum occurs at", cur_x)
    
    
    




from socialAD.forward import *
import numpy as np


### Implementation of Broyden's Optimization Method ###
# More computationally efficient than Newton's method #

#Broyden function
def broyden(function, variables, init_values):

	#Check user input types (string or list of strings)
	if not isinstance(function,str):

		#Vector function case
		if not isinstance(function,list):
			raise TypeError('function input must be either string or list of strings')

		if not isinstance(function[0],str):
			raise TypeError('function input must be either string or list of strings')

	#Check user input types (string or list of strings)
	if not isinstance(variables,str):

		#Vector function case
		if not isinstance(variables,list):
			raise TypeError('Each input must be either string or list of strings')

		if not isinstance(variables[0],str):
			raise TypeError('Each input must be either string or list of strings')

	#Check inital values is a list
	if not isinstance(init_values,list):
		
		raise TypeError('initial values must be a list')

	#Check that initial values and variables match in length
	if not len(variables) == len(init_values):

		raise Error('number of initial values must match number of variables')

	#Broyden's method for non-vector function
	if isinstance(function,str):

		#If just one string, convert to list
		if isinstance(variables,str):

			variables = [variables]

		#Convert to numpy arrays
		variables = np.array(variables)
		init_values = np.array(init_values)

		#Loop through variable names
		for index, variable in enumerate(variables):

			#Store to variable name
			name = variable

			#Save as forwardAD object
			exec(name + " = forwardAD(np.array([init_values[index]]), numvar = len(variables), idx = index)")

		#Stores full function as a forwardAD object
		AD_function = eval(function)

		#First iteration is Newton's method
		new_values = init_values - (AD_function.val / AD_function.der)

		#Store init as last values
		previous_values = init_values

		#Stores function value as previoust function value
		previous_function_val = AD_function.val

		#Iterate using finite differences until find good enough root
		while abs(AD_function.val) > 1e-7:

			#Loop through variable names
			for index, variable in enumerate(variables):

				#Store to variable name
				name = variable

				#Save as forwardAD object
				exec(name + " = forwardAD(np.array([new_values[index]]), numvar = len(variables), idx = index)")

			#Stores full function as a forwardAD object
			AD_function = eval(function)

			#Evaluate finite difference
			finite_difference = (AD_function.val - previous_function_val) / (new_values - previous_values)

			#Store function value
			previous_function_val = AD_function.val

			#Store domain values
			previous_values = new_values

			#Next estiamte
			new_values = previous_values - (AD_function.val / finite_difference)

		return new_values, AD_function.val

	#Case of vector function
	else:

		#List to store functions
		vector = []

		#If just one string, convert to list
		if isinstance(variables,str):

			variables = [variables]

		#Convert to numpy arrays
		variables = np.array(variables)
		init_values = np.array(init_values)

		#Loop through functions
		for func in function:

			#Loop through variable names
			for index, variable in enumerate(variables):

				#Store to variable name
				name = variable

				#Save as forwardAD object
				exec(name + " = forwardAD(np.array([init_values[index]]), numvar = len(variables), idx = index)")

			#Stores full function in vector of functions as forwardAD object
			vector.append(eval(func))

		#Store as a vector function
		AD_vector = vector_func(*vector)

		#Store values
		previous_values = init_values
		previous_values = previous_values.reshape(len(previous_values),1)

		#Store jacobian
		previous_jacobian = AD_vector.jacobian()

		#Store function values
		previous_function_val = AD_vector.values()

		new_values = previous_values - np.matmul(np.linalg.pinv(AD_vector.jacobian()),AD_vector.values())
		new_values = new_values.reshape(1,-1)[0]

		#Loop counter
		loop = 0

		#Iterate using finite differences until find good enough root
		while (np.linalg.norm(AD_vector.values()) > 1e-7):

			#List to store functions
			vector = []

			#Loop through functions
			for func in function:

				#Loop through variable names
				for index, variable in enumerate(variables):

					#Store to variable name
					name = variable

					#Save as forwardAD object
					exec(name + " = forwardAD(np.array([new_values[index]]), numvar = len(variables), idx = index)")

				#Stores full function in vector of functions as forwardAD object
				vector.append(eval(func))

			#Store as a vector function
			AD_vector = vector_func(*vector)

			#differences
			previous_values = previous_values.reshape(1,-1)[0]
			delta_f = AD_vector.values() - previous_function_val
			delta_x = new_values - previous_values
			# print(previous_values)
			delta_x = delta_x.reshape(len(delta_x),1)

			#Evaluate finite jacobian
			finite_jacobian = previous_jacobian + np.matmul((delta_f - np.matmul(previous_jacobian, delta_x)) / np.linalg.norm(delta_x),delta_x.T)

			#Store function value
			previous_function_val = AD_vector.values()

			#Store domain values
			previous_values = new_values
			previous_values = previous_values.reshape(len(previous_values),1)

			#Next estiamte
			new_values = previous_values - np.matmul(np.linalg.pinv(finite_jacobian),AD_vector.values())
			new_values = new_values.reshape(1,-1)[0]

			if loop > 10000:

				raise ValueError('Too many iterations - could not converge')

			loop += 1

		return new_values, AD_vector.values()


### Implementation of Newton's Method ###

#Newton function
def newton(function, variables, init_values):

	#Check user input types (string or list of strings)
	if not isinstance(function,str):

		#Vector function case
		if not isinstance(function,list):
			raise TypeError('function input must be either string or list of strings')

		if not isinstance(function[0],str):
			raise TypeError('function input must be either string or list of strings')

	#Check user input types (string or list of strings)
	if not isinstance(variables,str):

		#Vector function case
		if not isinstance(variables,list):
			raise TypeError('Each input must be either string or list of strings')

		if not isinstance(variables[0],str):
			raise TypeError('Each input must be either string or list of strings')

	#Check inital values is a list
	if not isinstance(init_values,list):
		
		raise TypeError('initial values must be a list')

	#Check that initial values and variables match in length
	if not len(variables) == len(init_values):

		raise Error('number of initial values must match number of variables')

	#Broyden's method for non-vector function
	if isinstance(function,str):

		#If just one string, convert to list
		if isinstance(variables,str):

			variables = [variables]

		#Convert to numpy arrays
		variables = np.array(variables)
		init_values = np.array(init_values)

		#Loop through variable names
		for index, variable in enumerate(variables):

			#Store to variable name
			name = variable

			#Save as forwardAD object
			exec(name + " = forwardAD(np.array([init_values[index]]), numvar = len(variables), idx = index)")

		#Stores full function as a forwardAD object
		AD_function = eval(function)

		#First iteration is Newton's method
		new_values = init_values - (AD_function.val / AD_function.der)

		#Store init as last values
		previous_values = init_values

		#Stores function value as previoust function value
		previous_function_val = AD_function.val

		#Iterate using finite differences until find good enough root
		while abs(AD_function.val) > 1e-7:

			#Loop through variable names
			for index, variable in enumerate(variables):

				#Store to variable name
				name = variable

				#Save as forwardAD object
				exec(name + " = forwardAD(np.array([new_values[index]]), numvar = len(variables), idx = index)")

			#Stores full function as a forwardAD object
			AD_function = eval(function)

			#Store function value
			previous_function_val = AD_function.val

			#Store domain values
			previous_values = new_values

			#Next estiamte
			new_values = previous_values - (AD_function.val / AD_function.der)

		return new_values, AD_function.val

	#Case of vector function
	else:

		#List to store functions
		vector = []

		#If just one string, convert to list
		if isinstance(variables,str):

			variables = [variables]

		#Convert to numpy arrays
		variables = np.array(variables)
		init_values = np.array(init_values)

		#Loop through functions
		for func in function:

			#Loop through variable names
			for index, variable in enumerate(variables):

				#Store to variable name
				name = variable

				#Save as forwardAD object
				exec(name + " = forwardAD(np.array([init_values[index]]), numvar = len(variables), idx = index)")

			#Stores full function in vector of functions as forwardAD object
			vector.append(eval(func))

		#Store as a vector function
		AD_vector = vector_func(*vector)

		#Store values
		previous_values = init_values
		previous_values = previous_values.reshape(len(previous_values),1)

		#Store jacobian
		previous_jacobian = AD_vector.jacobian()

		#Store function values
		previous_function_val = AD_vector.values()

		new_values = previous_values - np.matmul(np.linalg.pinv(AD_vector.jacobian()),AD_vector.values())
		new_values = new_values.reshape(1,-1)[0]

		#Loop counter
		loop = 0

		#Iterate using finite differences until find good enough root
		while (np.linalg.norm(AD_vector.values()) > 1e-7):

			#List to store functions
			vector = []

			#Loop through functions
			for func in function:

				#Loop through variable names
				for index, variable in enumerate(variables):

					#Store to variable name
					name = variable

					#Save as forwardAD object
					exec(name + " = forwardAD(np.array([new_values[index]]), numvar = len(variables), idx = index)")

				#Stores full function in vector of functions as forwardAD object
				vector.append(eval(func))

			#Store as a vector function
			AD_vector = vector_func(*vector)

			#Store function value
			previous_function_val = AD_vector.values()

			#Store domain values
			previous_values = new_values
			previous_values = previous_values.reshape(len(previous_values),1)

			#Next estiamte
			new_values = previous_values - np.matmul(np.linalg.pinv(AD_vector.jacobian()),AD_vector.values())
			new_values = new_values.reshape(1,-1)[0]

			if loop > 10000:

				raise ValueError('Too many iterations - could not converge')

			loop += 1

		return new_values, AD_vector.values()




import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()


setuptools.setup(
	name="socialAD",
	version="0.1.17",
	author="Ju Chulakadabba, Tao Tsui, Jenny Wang, Dash Young-Saver",
	author_email="ywang1@hsph.harvard.edu",
	description="Automatic differentiation",
	long_description_content_type="text/markdown",
	url="https://github.com/climate-change-is-real-python-dev/cs107-FinalProject",
	packages=setuptools.find_packages(),
	classifiers=[
	"Programming Language :: Python :: 3",
	"License :: OSI Approved :: MIT License",
	],
)
